'''
该模块用于global forecast system数据下载

如非特指，默认下载精度为0.25度。

数据说明: [https://www.ncei.noaa.gov/products/weather-climate-models/global-forecast](https://www.ncei.noaa.gov/products/weather-climate-models/global-forecast)

- `search` - 返回符合搜索条件的dem数据下载地址列表
- `download` - 将列表中的数据下载到本地
'''


from .downloader import download_sigletasking
import os
import subprocess


def get_gfs_from_ncep(date:str,creation_time:str,forecast_time:int,bbox=[115,38,136,54],save_dir='.',cover=False):
    '''
    从ncep官网下载近期gfs数据
    
    官方链接: [https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl](https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl)
    
    下载url示例: [https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl?file=gfs.t00z.pgrb2.0p25.f001&all_lev=on&all_var=on&subregion=&leftlon=115&rightlon=136&toplat=54&bottomlat=38&dir=%2Fgfs.20220815%2F00%2Fatmos](https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl?file=gfs.t00z.pgrb2.0p25.f001&all_lev=on&all_var=on&subregion=&leftlon=115&rightlon=136&toplat=54&bottomlat=38&dir=%2Fgfs.20220815%2F00%2Fatmos)

    Args:
        date (str): 下载日期，格式为：YYYYMMDD
        creation_time (str): 数据创建时间，可以是00、06、12、18中的一个
        forecast_time (int): 预测序列，范围1-384
        bbox (list): 下载数据的矩形范围
        save_dir (str): 存储文件夹
        cover (bool): 若文件已存在，是否覆盖

    '''

    # url_='https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25.pl?file=gfs.tCCz.pgrb2.0p25.fFFF&all_lev=on&all_var=on&subregion=&leftlon='+bbox[0]+'&rightlon='+bbox[2]+'&toplat='+bbox[3]+'&bottomlat='+bbox[1]+'&dir=%2Fgfs.YYYYMMDD%2FCC%2Fatmos'
    url_='https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25.pl?file=gfs.tCCz.pgrb2.0p25.fFFF&lev_10_m_above_ground=on&lev_2_m_above_ground=on&lev_entire_atmosphere=on&lev_entire_atmosphere_%5C%28considered_as_a_single_layer%5C%29=on&lev_surface=on&var_APCP=on&var_DSWRF=on&var_PWAT=on&var_RH=on&var_SPFH=on&var_TCDC=on&var_TMP=on&var_UGRD=on&var_VGRD=on&subregion=&leftlon='+str(bbox[0])+'&rightlon='+str(bbox[2])+'&toplat='+str(bbox[3])+'&bottomlat='+str(bbox[1])+'&dir=%2Fgfs.YYYYMMDD%2FCC%2Fatmos'
    file_name_='gfsYYYYMMDD.tCCz.pgrb2.0p25.fFFF'
    
    url=url_.replace('YYYYMMDD',date).replace('CC',creation_time.zfill(2)).replace('FFF',str(forecast_time).zfill(3))
    file_name=file_name_.replace('YYYYMMDD',date).replace('CC',creation_time.zfill(2)).replace('FFF',str(forecast_time).zfill(3))

    file_path=os.path.join(save_dir,file_name)
    if os.path.exists(file_path) and not cover:
        print(file_name + '已存在.')
        return

    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))
    
    download_sigletasking(url,os.path.join(save_dir,file_name))


def get_gfs_from_aws(date:str,creation_time:str,forecast_time:int,save_dir='.',cover=False):
    '''
    从aws下载gfs数据，数据时间范围从2021年2月26日开始。

    需要安装[aws cli](https://docs.aws.amazon.com/zh_cn/cli/latest/userguide/getting-started-install.html)
    
    建议使用aws cli命令行下载

    aws链接: [https://registry.opendata.aws/noaa-gfs-bdp-pds/](https://registry.opendata.aws/noaa-gfs-bdp-pds/)
    
    下载url示例: [https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl?file=gfs.t00z.pgrb2.0p25.f001&all_lev=on&all_var=on&subregion=&leftlon=115&rightlon=136&toplat=54&bottomlat=38&dir=%2Fgfs.20220815%2F00%2Fatmos](https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl?file=gfs.t00z.pgrb2.0p25.f001&all_lev=on&all_var=on&subregion=&leftlon=115&rightlon=136&toplat=54&bottomlat=38&dir=%2Fgfs.20220815%2F00%2Fatmos)

    Args:
        date (str): 下载日期，格式为：YYYYMMDD
        creation_time (str): 数据创建时间，可以是00、06、12、18中的一个
        forecast_time (int): 预测序列，范围1-384
        save_dir (str): 存储文件夹
        cover (bool): 若文件已存在，是否覆盖

    '''

    url_ = 's3://noaa-gfs-bdp-pds/gfs.YYYYMMDD/CC/atmos/gfs.tCCz.pgrb2.0p25.fFFF'
    if (date+creation_time.zfill(2)) < '2021032212' :
        url_ = 's3://noaa-gfs-bdp-pds/gfs.YYYYMMDD/CC/gfs.tCCz.pgrb2.0p25.fFFF'
    file_name_='gfsYYYYMMDD.tCCz.pgrb2.0p25.fFFF'

    url=url_.replace('YYYYMMDD',date).replace('CC',creation_time.zfill(2)).replace('FFF',str(forecast_time).zfill(3))
    file_name=file_name_.replace('YYYYMMDD',date).replace('CC',creation_time.zfill(2)).replace('FFF',str(forecast_time).zfill(3))
    
    file_path=os.path.join(save_dir,file_name)
    if os.path.exists(file_path) and not cover:
        print(file_name + '已存在.')
        return

    if not os.path.exists(os.path.dirname(file_path)):
                os.makedirs(os.path.dirname(file_path))
    
    # download_sigletasking(url,os.path.join(save_dir,file_name)
    subprocess.call(['aws', 's3', 'cp', url, file_path])


def get_gfs_from_gee(date:str,creation_time:str,forecast_time:int,bbox=[115,38,136,54],save_dir='.',cover=False):
    '''
    从google earth engine下载获取gfs数据，时间范围从2015年7月1日开始。

    需要科学上网！
    
    需要gee for python本地配置
    - pip install google-api-python-client
    - pip install pyCrytod
    - pip install earthengine-api
    - earthengine authenticate (需要gcloud)

    gfs from gee链接: [https://developers.google.cn/earth-engine/datasets/catalog/NOAA_GFS0P25](https://developers.google.cn/earth-engine/datasets/catalog/NOAA_GFS0P25)
    
    Args:
        date (str): 下载日期，格式为：YYYYMMDD
        creation_time (str): 数据创建时间，可以是00、06、12、18中的一个
        forecast_time (int): 预测序列，范围1-384
        bbox (list): 下载数据的矩形范围
        save_dir (str): 存储文件夹
        cover (bool): 若文件已存在，是否覆盖

    '''
    pass


"""Main module."""
import pathlib
import os
import s3fs

minio_paras = {
    'endpoint_url' : 'http://minio.waterism.com:9000',
    'access_key' : '',
    'secret_key' : '',
    'bucket_name' : 'test'
}

home_path = str(pathlib.Path.home())

if os.path.exists(os.path.join(home_path,'.wisminio')):
    for line in open(os.path.join(home_path,'.wisminio')):
        key = line.split('=')[0].strip()
        value = line.split('=')[1].strip()
        # print(key,value)
        if key == 'endpoint_url':
            minio_paras['endpoint_url'] = value
        elif key == 'access_key':
            minio_paras['access_key'] = value
        elif key == 'secret_key':
            minio_paras['secret_key'] = value
        elif key == 'bucket_path':
            minio_paras['bucket_name'] = value
            
fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": minio_paras['endpoint_url']}, 
    key=minio_paras['access_key'], 
    secret=minio_paras['secret_key'] 
)


ro = {
    'client_kwargs': {'endpoint_url': minio_paras['endpoint_url']}, 
    'key': minio_paras['access_key'], 
    'secret': minio_paras['secret_key']
}


"""Top-level package for hydro_opendata."""
import pathlib
import os

from .common import minio_paras


__author__ = """Jianfeng Zhu"""
__email__ = 'zjf014@gmail.com'
__version__ = '0.0.1'

home_path = str(pathlib.Path.home())

if os.path.exists(os.path.join(home_path,'.wisminio')):
    pass
else:
    
    access_key = input('minio_access_key:')
    secret_key = input('minio_secret_key:')

    f = open(os.path.join(home_path,'.wisminio'),'w')
    f.write('endpoint_url = ' + minio_paras['endpoint_url'])
    f.write('\naccess_key = ' + access_key)
    f.write('\nsecret_key = ' + secret_key)
    f.write('\nbucket_name = ' + minio_paras['bucket_name'])
    f.close()


import sys
sys.path.append('/home/zhujianfeng/wis/hydro-opendata/')

# print(sys.path)

print('----------------------------------------------------------------------------')
# from hydro_opendata.stac.minio import Era5_land, GPM_IMERG_Early, GFS_atmos
import geopandas as gpd

aoi = gpd.read_file('test.geojson')
print(aoi)
print()

# era5 = Era5_land()
# e = era5.search(aoi=aoi)
# print(e)
# print()

# gpm = GPM_IMERG_Early()
# g = gpm.search(aoi=aoi)
# print(g)
# print()

# gfs = GFS_atmos('tp')
# f = gfs.search(aoi=aoi)
# print(f)
# print()

# print('----------------------------------------------------------------------------')
from hydro_opendata.s3api.minio import ERA5LReader
import numpy as np

era5 = ERA5LReader()

# 指定开始及结束时间
start_time=np.datetime64("2021-06-01T00:00:00.000000000")
end_time=np.datetime64("2021-06-30T23:00:00.000000000")

# 通过指定四至范围读取
bbox=(121,39,123,40)
ds1 = era5.open_dataset(data_variables=['Total precipitation'], start_time=start_time, end_time=end_time, bbox=bbox)
print(ds1)
# print()

# gr = GPMReader('wis')
# ds2 = gfs.open_dataset(creation_date=np.datetime64("2022-08-01"))
# print(ds2)
# print()

# start_time=np.datetime64("2023-08-23T00:00:00.000000000")
# end_time=np.datetime64("2023-08-24T23:30:00.000000000")
# ds3 = gpm.from_aoi(start_time=start_time, end_time=end_time, aoi=aoi)
# print(ds3)

# print('----------------------------------------------------------------------------')
# from hydro_opendata.downloader.dem import Alos_DEM
# dem = Alos_DEM()
# # sources = dem.list_sources()
# # print(sources)

# # new = dem.add_source('aws','https:\\','dem')
# # print(new)
# # sources = dem.list_sources()
# # print(sources)
# hrefs = dem.search()
# print(len(hrefs))



# ds = gr.open_dataset(start_time=start_time, end_time=end_time, bbox=(134,52, 137,56))
# print(ds)

# print('----------------------------------------------------------------------------')

# from hydro_opendata.data.minio import Era5L, GPM, GFS

# era5 = Era5L('wis')
# print(era5.catalog.spatial_resolution)
# print(era5.catalog.start_time)
# print(era5.catalog.end_time)

# start_time=np.datetime64("2021-06-30T00:00:00")
# end_time=np.datetime64("2021-07-31T23:30:00")
# e = era5.catalog.search(aoi=aoi, start_time=start_time, end_time=end_time)
# print(e)


# eds = era5.reader.open_dataset(data_variables=['Total precipitation'], start_time=start_time, end_time=end_time, bbox=(121,39,122,40))
# print(eds)
# print()
# eds = era5.reader.from_shp(data_variables=['Total precipitation'], start_time=start_time, end_time=end_time, shp='test.geojson')
# print(eds)
# print()
# eds = era5.reader.from_aoi(data_variables=['Total precipitation'], start_time=start_time, end_time=end_time, aoi=aoi)
# print(eds)
# print()

# eds = era5.reader.to_netcdf(data_variables=['Total precipitation'], start_time=start_time, end_time=end_time, shp='test.geojson', resolution='6-hourly')
# print(eds)
# print()

# print('----------------------------------------------------------------------------')

# gpm = GPM('wis')
# print(gpm.catalog.spatial_resolution)
# print(gpm.catalog.start_time)
# print(gpm.catalog.end_time)

# start_time=np.datetime64("2023-06-30T00:00:00")
# end_time=np.datetime64("2023-07-31T23:30:00")
# g = gpm.catalog.search(aoi=aoi, start_time=start_time, end_time=end_time)
# print(g)

# gds = gpm.reader.open_dataset(start_time=start_time, end_time=end_time, bbox=(121,39,122,40))
# print(gds)
# print()
# gds = gpm.reader.from_shp(start_time=start_time, end_time=end_time, shp='test.geojson')
# print(gds)
# print()
# gds = gpm.reader.from_aoi(start_time=start_time, end_time=end_time, aoi=aoi)
# print(gds)
# print()

# print('----------------------------------------------------------------------------')

# gfs = GFS('wis', 'tp')
# print(gfs.catalog.spatial_resolution)
# print(gfs.catalog.start_time)
# print(gfs.catalog.end_time)

# g = gfs.catalog.search(aoi=aoi)
# print(g)

# creation_date = np.datetime64("2022-06-30")
# creation_time = '00'
# gds = gfs.reader.open_dataset(creation_date=creation_date, creation_time=creation_time, bbox=(121,39,122,40))
# print(gds)
# print()
# gds = gfs.reader.from_shp(creation_date=creation_date, creation_time=creation_time, shp='test.geojson')
# print(gds)
# print()
# gds = gfs.reader.from_aoi(creation_date=creation_date, creation_time=creation_time, aoi=aoi)
# print(gds)
# print()


'''
该模块用于获取minio服务器中的数据列表，包括：

- `Era5_land`
- `GPM_IMERG_Early`
- `GFS_atmos`
'''

from ..common import minio_paras, fs
import os
import numpy as np
import pandas as pd
import geopandas as gpd
import json

bucket_name = minio_paras['bucket_name']

class ERA5LCatalog():
    '''
    用于获取era5-land的数据源信息，并搜索minio服务器中的数据范围
    
    Attributes:
        collection_id (str): 数据集名称
        data_sources (str): 数据源
        description (str): 数据源链接
        spatial_resolution (str): 空间分辨率
        temporal_resolution (str): 时间分辨率
        
        start_time (str): minio服务器中数据起始时间
        end_time (str): minio服务器中数据结束时间
        bbox (str): minio服务器中数据空间范围
        
    Method:
        search(aoi, start_time, end_time): 搜索minio服务器中的数据范围
    '''
    
    def __init__(self):
        
        self._collection_id = 'era5-land'
        self._datasources = 'ECMWF'
        self._description = 'https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-land?tab=overview'
        self._spatialresolution = '0.1 x 0.1; Native resolution is 9 km.'
        self._temporalresolution = 'hourly'

        # with fs.open('test/geodata/era5_land.json/')
        self._starttime = np.datetime64('2015-07-01T00:00:00')
        self._endtime = np.datetime64('2021-12-31T23:00:00')
        self._bbox = (115, 38, 136, 54)

    @property
    def collection_id(self):
        return self._collection_id
    
    @property
    def data_sources(self):
        return self._datasources
    
    @property
    def description(self):
        return self._description
    
    @property
    def spatial_resolution(self):
        return self._spatialresolution
    
    @property
    def temporal_resolution(self):
        return self._temporalresolution
    
    @property
    def start_time(self):
        return self._starttime
    
    @property
    def end_time(self):
        return self._endtime
    
    @property
    def bbox(self):
        return self._bbox
        
    def search(self, aoi, start_time=None, end_time=None):
        '''
        查询并获取数据清单

        Args:
            aoi (GeoDataFrame): 矢量数据范围
            strt_time (datatime64): 查询的起始时间
            end_time (datatime64): 查询的终止时间

        Returns:
            datalist (GeoDataFrame): 符合条件的数据清单
        '''
        
        if start_time is None:
            start_time = self._starttime
        if end_time is None:
            end_time = self._endtime
        
        if start_time < end_time:
            if start_time < self._starttime:
                start_time = self._starttime
            if end_time > self._endtime:
                end_time = self._endtime
                
            df = pd.DataFrame(
                {
                    'id': [self._collection_id],
                    'start_time': [start_time],
                    'end_time': [end_time],
                    'geometry': ['POLYGON((115 54,115 38,136 38,136 54,115 54))']
                }
            )
            df['geometry'] = gpd.GeoSeries.from_wkt(df['geometry'])
            gdf = gpd.GeoDataFrame(df, geometry='geometry',crs='EPSG:4326')
            
            clip = gdf.clip(aoi)
            
            return clip
            
        else:
            return None
        

class GPMCatalog():
    '''
    用于获取gpm的数据源信息，并搜索minio服务器中的数据范围
    
    Attributes:
        collection_id (str): 数据集名称
        data_sources (str): 数据源
        description (str): 数据源链接
        spatial_resolution (str): 空间分辨率
        temporal_resolution (str): 时间分辨率
        
        start_time (str): minio服务器中数据起始时间
        end_time (str): minio服务器中数据结束时间
        bbox (str): minio服务器中数据空间范围
        
    Method:
        search(aoi, start_time, end_time): 搜索minio服务器中的数据范围
    '''
    
    def __init__(self):
        self._collection_id = 'gpm-imerg-early'
        self._datasources = 'NASA & JAXA'
        self._description = 'https://disc.gsfc.nasa.gov/datasets/GPM_3IMERGHHE_06/summary'
        self._spatialresolution = '0.1 x 0.1; Native resolution is 9 km. (60°S-60°N)'
        self._temporalresolution = 'half-hourly'

        with fs.open(os.path.join(bucket_name,'geodata/gpm/gpm.json')) as f:
            gpm = json.load(f)
            self._starttime = np.datetime64(gpm['start'])
            self._endtime = np.datetime64(gpm['end'])
            self._bbox = gpm['bbox']

    @property
    def collection_id(self):
        return self._collection_id
    
    @property
    def data_sources(self):
        return self._datasources
    
    @property
    def description(self):
        return self._description
    
    @property
    def spatial_resolution(self):
        return self._spatialresolution
    
    @property
    def temporal_resolution(self):
        return self._temporalresolution
    
    @property
    def start_time(self):
        return self._starttime
    
    @property
    def end_time(self):
        return self._endtime
    
    @property
    def bbox(self):
        return self._bbox            
            
    def search(self, aoi, start_time=None, end_time=None):
        '''
        查询并获取数据清单

        Args:
            aoi (GeoDataFrame): 矢量数据范围
            strt_time (datatime64): 查询的起始时间
            end_time (datatime64): 查询的终止时间

        Returns:
            datalist (GeoDataFrame): 符合条件的数据清单
        '''
        
        if start_time is None:
            start_time = self._starttime
        if end_time is None:
            end_time = self._endtime
            
        if start_time < end_time:
            if start_time < self._starttime:
                start_time = self._starttime
            if end_time > self._endtime:
                end_time = self._endtime
                
            df = pd.DataFrame(
                {
                    'id': [self._collection_id],
                    'start_time': [start_time],
                    'end_time': [end_time],
                    'geometry': [f'POLYGON(({self._bbox[0]} {self._bbox[3]},{self._bbox[0]} {self._bbox[1]},{self._bbox[2]} {self._bbox[1]},{self._bbox[2]} {self._bbox[3]},{self._bbox[0]} {self._bbox[3]}))']
                }
            )
            df['geometry'] = gpd.GeoSeries.from_wkt(df['geometry'])
            gdf = gpd.GeoDataFrame(df, geometry='geometry',crs='EPSG:4326')
            
            clip = gdf.clip(aoi)
            
            return clip
            
        else:
            return None
        
class GFSCatalog():
    '''
    用于获取gfs的数据源信息，并搜索minio服务器中的数据范围
    
    Attributes:
        collection_id (str): 数据集名称
        data_sources (str): 数据源
        description (str): 数据源链接
        spatial_resolution (str): 空间分辨率
        temporal_resolution (str): 时间分辨率
        
        start_time (str): minio服务器中数据起始时间
        end_time (str): minio服务器中数据结束时间
        
    Method:
        search(aoi, start_time, end_time): 搜索minio服务器中的数据范围
    '''
    
    def __init__(self, variable='tp'):
        self._variable = variable
        self._collection_id = f'gfs_atmos.{variable}'
        self._datasources = 'NOAA'
        self._description = 'https://www.emc.ncep.noaa.gov/emc/pages/numerical_forecast_systems/gfs.php'
        self._spatialresolution = '0.25 x 0.25'
        self._temporalresolution = 'hourly; 1-120h'

        with fs.open(os.path.join(bucket_name,'geodata/gfs/gfs.json')) as f:
            gfs = json.load(f)
            self._starttime = np.datetime64(gfs[variable][0]['start'])
            self._endtime = np.datetime64(gfs[variable][-1]['end'])
            # self.bbox = gpm['bbox']
            
    @property
    def variable(self):
        return self._variable    
    
    @property
    def collection_id(self):
        return self._collection_id
    
    @property
    def data_sources(self):
        return self._datasources
    
    @property
    def description(self):
        return self._description
    
    @property
    def spatial_resolution(self):
        return self._spatialresolution
    
    @property
    def temporal_resolution(self):
        return self._temporalresolution
    
    @property
    def start_time(self):
        return self._starttime
    
    @property
    def end_time(self):
        return self._endtime
    
    def search(self, aoi, start_time=None, end_time=None):
        '''
        查询并获取数据清单

        Args:
            aoi (GeoDataFrame): 矢量数据范围
            start_time (datatime64): 查询的起始时间
            end_time (datatime64): 查询的终止时间

        Returns:
            datalist (GeoDataFrame): 符合条件的数据清单
        '''
        
        if start_time is None:
            start_time = self._starttime
        if end_time is None:
            end_time = self._endtime
        
        if start_time < end_time:
            
            stime = []
            etime = []
            bbox = []
            ids = []
            
            with fs.open(os.path.join(bucket_name,'geodata/gfs/gfs.json')) as f:
                cont = json.load(f)
                clist = cont[self._variable]
            
            for c in clist:

                if start_time < np.datetime64(c['start']):
                    stime.append(np.datetime64(c['start']))
                else:
                    stime.append(start_time)

                if end_time > np.datetime64(c['end']):
                    etime.append(np.datetime64(c['end']))
                else:
                    etime.append(end_time)

                bb = c['bbox']
                bbox.append(f'POLYGON(({bb[0]} {bb[3]},{bb[0]} {bb[1]},{bb[2]} {bb[1]},{bb[2]} {bb[3]},{bb[0]} {bb[3]}))')
                ids.append(self._collection_id)

            df = pd.DataFrame(
                {
                    'id': ids,
                    'start_time': stime,
                    'end_time': etime,
                    'geometry': bbox
                }
            )
            df['geometry'] = gpd.GeoSeries.from_wkt(df['geometry'])
            gdf = gpd.GeoDataFrame(df, geometry='geometry',crs='EPSG:4326')
            
            clip = gdf.clip(aoi)
            
            return clip
            
        else:
            return None


'''
该模块用于alos-dem数据下载

数据源: [Microsoft Planetary Computer](https://planetarycomputer.microsoft.com/dataset/alos-dem)

数据目录: [https://planetarycomputer.microsoft.com/api/stac/v1/](https://planetarycomputer.microsoft.com/api/stac/v1/)

数据说明: [https://www.eorc.jaxa.jp/ALOS/en/aw3d30/aw3d30v3.2_product_e_e1.2.pdf](https://www.eorc.jaxa.jp/ALOS/en/aw3d30/aw3d30v3.2_product_e_e1.2.pdf)

'''

# from pystac_client import Client
import requests
import os
import json

from .downloader import download_by_stream, download_sigletasking


class Alos_DEM:

    def __init__(self):
        self._sources = {
            'microsoft':{'url':'https://planetarycomputer.microsoft.com/api/stac/v1/','collection_name':'alos-dem'},
        }
        self._hrefs=[]
        
    def list_sources(self):
        '''
        返回stac列表

        Returns:
            sources (list): stac列表
        '''   
        return self._sources
    
    def add_source(self, source_name, url, collection_name):
        '''
        添加自定义stac内容

        Args:
            source_name (str): 源名称
            url (list): STAC根路径地址
            collection_name (str): 数据集名称

        Returns:
            source (dict): 符合条件的url地址列表
        '''
        
        new = {source_name:{'url':url, 'collection_name':collection_name}}
        self._sources[source_name] = {'url':url, 'collection_name':collection_name}
        
        return new

    def search(self, source='microsoft', bbox=[115,38,136,54], intersects=None):
        '''
        搜索符合条件的dem，并返回url地址列表

        Args:
            catalog_url (str): STAC根路径地址，默认为[微软planetarycomputer](https://planetarycomputer.microsoft.com)
            bbox (list): 搜索的矩形范围
            intersects (str): 搜索范围，应符合GeoJSON的geometry属性格式

        Returns:
            hrefs (list): 符合条件的url地址列表
        '''
        
        if source not in self._sources.keys():
            return
        
        stac =  self._sources[source]['url']
        stac_response = requests.get(stac).json()
        catalog_links = stac_response['links']
        search = [l['href'] for l in catalog_links if l['rel'] == 'search'][0]
        
        params = {}
        params['collections'] = [self._sources[source]['collection_name']]
                
        if bbox is not None:
            params['bbox'] = bbox
        if intersects is not None:
            params['intersects'] = intersects
        
        query = requests.post(search, json=params).json()
        
        features = query['features']
        
        
        
        for feat in features:
            self._hrefs.append(feat['assets']['data']['href'])

        return self._hrefs


    def download(self, save_dir='.', cover=False):
        '''
        下载列表中的dem数据

        Args:
            save_dir (str): 本地保存目录
            cover (bool): 若文件已存在，是否覆盖

        '''

        if hrefs is not None:

            for href in hrefs:
                url=requests.utils.urlparse(href)
                file_path=os.path.join(save_dir,url.path.split('/')[-1])

                if os.path.exists(file_path) and not cover:
                    continue

                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))

                download_sigletasking(href,file_path)


'''
该模块用于从minio中读取gpm数据，主要方法包括：

- `open_dataset` - 普通读取数据方法
- `from_shp` - 通过已有矢量范围读取数据方法
- `from_aoi` - 通过已有GeoDataFrame范围读取数据方法
'''

import os
import numpy as np
import s3fs
import xarray as xr
import calendar
import dask
import json

from ..common import minio_paras, fs, ro
from ..utils import regen_box

bucket_name = minio_paras['bucket_name']


# 后期从minio读取
start = np.datetime64("2016-01-01T00:00:00.000000000")
end = np.datetime64("2023-08-17T23:30:00.000000000")
change = np.datetime64("2023-07-01T23:30:00.000000000")

with fs.open(os.path.join(bucket_name, 'geodata/gpm/gpm.json')) as f:
    cont = json.load(f)
end = np.datetime64(cont['end'])

box = (73.05, 3.05, 135.95, 53.95)

variables=[
    'HQobservationTime',
    'HQprecipSource',
    'HQprecipitation',
    'IRkalmanFilterWeight',
    'IRprecipitation',
    'precipitationCal',
    'precipitationUncal',
    'probabilityLiquidPrecipitation',
    'randomError'
]


dask.config.set({"array.slicing.split_large_chunks": False})

def get_dataset_year(start_time, end_time, bbox, time_chunks):
    
    year = str(start_time)[:4]
    
    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://", 
        engine="zarr", 
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": fs.open(os.path.join('s3://' + bucket_name , f'geodata/gpm/{year}/gpm{year}_inc.json')), 
                "remote_protocol": "s3",
                "remote_options": ro
            }
        }      
    )
    
    ds = ds['precipitationCal']
    # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()
    
    # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose('time','lon','lat')
    
    if start_time < start:
        start_time = start
    
    if end_time > end:
        end_time = end
    
    times = slice(start_time, end_time)
    ds = ds.sel(time=times)
    
    bbox = regen_box(bbox, 0.1, 0.05)
    
    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]
        
    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]
    
    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]
    
    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]
    
    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)
    
    ds = ds.sortby('lat', ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)
    
    return ds

def get_dataset_month(start_time, end_time, bbox, time_chunks):
    
    year = str(start_time)[:4]
    month = str(start_time)[5:7].zfill(2)
    
    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://", 
        engine="zarr", 
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": fs.open(os.path.join('s3://' + bucket_name , f'geodata/gpm/{year}/{month}/gpm{year}{month}_inc.json')), 
                "remote_protocol": "s3",
                "remote_options": ro
            }
        }      
    )
    
    ds = cf2datetime(ds)
    ds = ds['precipitationCal']
    # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()
    
    # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose('time','lon','lat')
    
    if start_time < start:
        start_time = start
    
    if end_time > end:
        end_time = end
    
    times = slice(start_time, end_time)
    ds = ds.sel(time=times)
    
    bbox = regen_box(bbox, 0.1, 0.05)
    
    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]
        
    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]
    
    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]
    
    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]
    
    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)
    
    ds = ds.sortby('lat', ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)
    
    return ds

def get_dataset_day(start_time, end_time, bbox, time_chunks):
    
    year = str(start_time)[:4]
    month = str(start_time)[5:7].zfill(2)
    day = str(end)[8:10].zfill(2)
    
    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://", 
        engine="zarr", 
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": fs.open(os.path.join('s3://' + bucket_name, f'geodata/gpm/{year}/{month}/gpm{year}{month}_{day}.json')), 
                "remote_protocol": "s3",
                "remote_options": ro
            }
        }      
    )
    
    ds = cf2datetime(ds)
    ds = ds['precipitationCal']
    # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()
    
    # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose('time','lon','lat')
    
    if start_time < start:
        start_time = start
    
    if end_time > end:
        end_time = end
    
    times = slice(start_time, end_time)
    ds = ds.sel(time=times)
    
    bbox = regen_box(bbox, 0.1, 0.05)
    
    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]
        
    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]
    
    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]
    
    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]
    
    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)
    
    ds = ds.sortby('lat', ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)
    
    return ds

def cf2datetime(ds):
    ds = ds.copy()
    time_tmp1 = ds.indexes['time']
    attrs = ds.coords['time'].attrs
    time_tmp2 = []
    for i in range(time_tmp1.shape[0]):
        tmp = time_tmp1[i]
        a = str(tmp.year).zfill(4)
        b = str(tmp.month).zfill(2)
        c = str(tmp.day).zfill(2)
        d = str(tmp.hour).zfill(2)
        e = str(tmp.minute).zfill(2)
        f = str(tmp.second).zfill(2)
        time_tmp2.append(np.datetime64('{}-{}-{} {}:{}:{}.00000000'.format(a,b,c,d,e,f))) 
    ds = ds.assign_coords(time = time_tmp2)
    ds.coords['time'].attrs = attrs
    
    return ds

def open_dataset(start_time=np.datetime64("2023-01-01T00:00:00.000000000"), end_time=np.datetime64("2023-01-02T00:00:00.000000000"), bbox=box, time_chunks=48):
    '''
    从minio服务器读取gpm数据

    Args:
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        bbox (list|tuple): 四至范围
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    '''
    
    if end_time <= start_time:
        raise Exception('结束时间不能早于开始时间')
        
    if end_time <= change:
        # 早于20230701
        
        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])

        if year_start == year_end:

            ds = get_dataset_year(start_time=start_time, end_time=end_time, bbox=bbox, time_chunks=time_chunks)
            return ds

        elif year_start < year_end:

            dss = []
            years = range(year_start, year_end + 1)
            for year in years:

                if year == year_start:

                    dss.append(get_dataset_year(start_time=start_time, end_time=np.datetime64(f"{year}-12-31T23:30:00.000000000"), bbox=bbox, time_chunks=time_chunks))

                elif year == year_end:

                    dss.append(get_dataset_year(start_time=np.datetime64(f"{year}-01-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))

                else:
                    dss.append(
                        get_dataset_year(
                            start_time=np.datetime64(f"{year}-01-01T00:00:00.000000000"), 
                            end_time=np.datetime64(f"{year}-12-31T23:30:00.000000000"), 
                            bbox=bbox, 
                            time_chunks=time_chunks
                        )
                    )
            return xr.merge(dss)
    
    elif change <= start_time:
        # 晚于20230701
        
        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])
        month_start = int(str(start_time)[5:7])
        month_end = int(str(end_time)[5:7])
        end_month = int(str(end)[5:7])
                
        if year_start == year_end:
            
            if month_end < end_month:
                
                if month_start == month_end:
                    ds = get_dataset_month(start_time=start_time, end_time=end_time, bbox=bbox, time_chunks=time_chunks)
                    return ds
                
                else:
                    dss = []
                    for m in range(month_start, month_end + 1):
                        if m == month_start:
                            d = calendar.monthrange(year_start,m)[1]
                            dss.append(get_dataset_month(start_time=start_time, end_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), bbox=bbox, time_chunks=time_chunks))
                        elif m == month_end:
                            dss.append(get_dataset_month(start_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))
                        else:
                            d = calendar.monthrange(year_start,m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                    end_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                    bbox=bbox, 
                                    time_chunks=time_chunks
                                )
                            )
                            
                    return xr.merge(dss)
        
            else:
                
                if month_start == month_end:
                    ds = get_dataset_day(start_time=start_time, end_time=end_time, bbox=bbox, time_chunks=time_chunks)
                    return ds
                
                else:
                    dss = []
                    for m in range(month_start, month_end + 1):
                        if m == month_start:
                            d = calendar.monthrange(year_start,m)[1]
                            dss.append(get_dataset_month(start_time=start_time, end_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), bbox=bbox, time_chunks=time_chunks))
                        elif m == month_end:
                            dss.append(get_dataset_day(start_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))
                        else:
                            d = calendar.monthrange(year_start,m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                    end_time=np.datetime64(f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                    bbox=bbox, 
                                    time_chunks=time_chunks
                                )
                            )
                            
                    return xr.merge(dss)
                
        else:
            
            dss = []
            
            for y in range(year_start, year_end):
                
                if y == year_start:
                    
                    for m in range(month_start,13):
                        if m == month_start:
                            d = calendar.monthrange(y,m)[1]
                            dss.append(get_dataset_month(start_time=start_time, end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), bbox=bbox, time_chunks=time_chunks))
                        else:
                            d = calendar.monthrange(y,m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                    end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                    bbox=bbox, 
                                    time_chunks=time_chunks
                                )
                            )
                
                elif y == year_end:
                    
                    for m in range(1, month_end + 1):
                        if m == end_month:
                            dss.append(get_dataset_day(start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))
                        else:
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                    end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                    bbox=bbox, 
                                    time_chunks=time_chunks
                                )
                            )
                
                else:
                    for m in range(1, 13):
                        d = calendar.monthrange(y,m)[1]
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                bbox=bbox, 
                                time_chunks=time_chunks
                            )
                        )

            return xr.merge(dss)
        
    else:
        # 包含20230701
        
        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])
        month_start = int(str(start_time)[5:7])
        month_end = int(str(end_time)[5:7])
        end_month = int(str(end)[5:7])
        
        dss = []
        
        for y in range(year_start, 2024):
            if y == year_start and y == 2023:
                dss.append(
                    get_dataset_year(
                        start_time=start_time, 
                        end_time=np.datetime64("2023-07-01T23:30:00.000000000"), 
                        bbox=bbox, 
                        time_chunks=time_chunks
                    )
                )
            elif y == year_start and y < 2023:
                dss.append(
                    get_dataset_year(
                        start_time=start_time, 
                        end_time=np.datetime64("2023-12-31T23:30:00.000000000"), 
                        bbox=bbox, 
                        time_chunks=time_chunks
                    )
                )
            elif y == 2023:
                dss.append(
                    get_dataset_year(
                        start_time=np.datetime64("2023-01-01T00:00:00.000000000"), 
                        end_time=np.datetime64("2023-07-01T23:30:00.000000000"), 
                        bbox=bbox, 
                        time_chunks=time_chunks
                    )
                )
            else:
                dss.append(
                    get_dataset_year(
                        start_time=np.datetime64("2023-01-01T00:00:00.000000000"), 
                        end_time=np.datetime64("2023-12-31T23:30:00.000000000"), 
                        bbox=bbox, 
                        time_chunks=time_chunks
                    )
                )
                
        for y in range(2023, year_end + 1):
            
            if y == year_end and y == 2023:
                for m in range(7, month_end + 1):
                    if m == month_end and m == 7:
                        dss.append(get_dataset_month(start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-02T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))                    
                    elif m == month_end and m == end_month:
                        dss.append(get_dataset_day(start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))
                    elif m == month_end and m > 7:
                        dss.append(get_dataset_month(start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))
                    elif m == 7:
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64("2023-07-02T00:00:00.000000000"), 
                                end_time=np.datetime64("2023-07-31T23:30:00.000000000"), 
                                bbox=bbox, 
                                time_chunks=time_chunks
                            )
                        )
                    else:
                        d = calendar.monthrange(y,m)[1]
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                bbox=bbox, 
                                time_chunks=time_chunks
                            )
                        )
            elif y == year_end and y > 2023:
                for m in range(1, month_end + 1):
                    if m == month_end and m == end_month:
                        dss.append(get_dataset_day(start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))                  
                    elif m == month_end:
                        dss.append(get_dataset_month(start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), end_time=end_time, bbox=bbox, time_chunks=time_chunks))
                    else:
                        d = calendar.monthrange(y,m)[1]
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                bbox=bbox, 
                                time_chunks=time_chunks
                            )
                        )
            elif y == 2023:
                for m in range(7,13):
                    if m == 7:
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64("2023-07-02T00:00:00.000000000"), 
                                end_time=np.datetime64("2023-07-31T23:30:00.000000000"), 
                                bbox=bbox, 
                                time_chunks=time_chunks
                            )
                        )
                    else:
                        d = calendar.monthrange(y,m)[1]
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                                end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                                bbox=bbox, 
                                time_chunks=time_chunks
                            )
                        )
            else:
                for m in range(1,13):
                    d = calendar.monthrange(y,m)[1]
                    dss.append(
                        get_dataset_year(
                            start_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"), 
                            end_time=np.datetime64(f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"), 
                            bbox=bbox, 
                            time_chunks=time_chunks
                        )
                    )
   
        return xr.merge(dss)

import geopandas as gpd

def from_shp(start_time=np.datetime64("2023-01-01T00:00:00.000000000"), end_time=np.datetime64("2023-01-02T00:00:00.000000000"), shp=None, time_chunks=48):
    '''
    通过已有的矢量数据范围从minio服务器读取gpm数据

    Args:
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        shp (str): 矢量数据路径
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    '''
    
    gdf = gpd.GeoDataFrame.from_file(shp)
    b = gdf.bounds
    bbox = regen_box((b.loc[0]['minx'],b.loc[0]['miny'],b.loc[0]['maxx'],b.loc[0]['maxy']), 0.1, 0.05)

    ds = open_dataset(start_time, end_time, bbox, time_chunks)
    
    return ds

def from_aoi(start_time=np.datetime64("2023-01-01T00:00:00.000000000"), end_time=np.datetime64("2023-01-02T00:00:00.000000000"), aoi:gpd.GeoDataFrame=None, time_chunks=48):
    '''
    用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gpm数据

    Args:
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    '''
    
    b = aoi.bounds
    bbox = regen_box((b.loc[0]['minx'],b.loc[0]['miny'],b.loc[0]['maxx'],b.loc[0]['maxy']), 0.1, 0.05)

    ds = open_dataset(start_time, end_time, bbox, time_chunks)
    
    return ds

if __name__ == '__main__':
    
    pass


'''
该模块用于下载，通常是url链接或执行命令

- `download_from_url` - 通过url链接下载文件
- `download_by_stream` - 通过stream方式下载url链接
- `download_singletasking` - 单函数线程下载文件，显示进度条
'''

import requests
from tqdm import tqdm

def download_from_url(url,file_name):
    '''
    通过url链接下载文件的一般方法

    Args:
        url (str): url链接
        file_name (str): 文件的本地存储地址

    '''
    
    f=requests.get(url)
    with open(file_name,"wb") as tiff:
            tiff.write(f.content)
    f.close()


def download_by_stream(url, file_name):
    '''
    通过stream下载url链接文件

    Args:
        url (str): url链接
        file_name (str): 文件的本地存储地址

    '''

    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36 QIHU 360SE'
    }

    # 发起 head 请求，即只会获取响应头部信息
    head = requests.head(url, headers=headers)

    
    if head.status_code == 404 :
        print("404 not found.")
    elif head.status_code == 200 :

        response = requests.get(url, headers=headers, stream=True)
        with open(file_name, mode='wb') as f:
            # 写入分块文件
            for i,chunk in enumerate(response.iter_content(chunk_size=1024),start=1):
                f.write(chunk)
                print('\r', "已下载：%.2f MB"%(i/1024), end='', flush=True)

            print(file_name+"下载完成。")
           
def download_sigletasking(url: str, file_name: str):
    '''
    单函数线程下载文件，显示进度条

    Args:
        url (str): 文件链接
        file_name (str): 文件名或文件路径

    '''
    # 文件下载直链
    # 请求头
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36 QIHU 360SE'
    }

    # 发起 head 请求，即只会获取响应头部信息
    head = requests.head(url, headers=headers)

    if head.status_code == 404 :
        print("404 not found.")
    elif head.status_code == 200 :

        # 文件大小，以 B 为单位
        file_size = head.headers.get('Content-Length')
        if file_size is not None:
            file_size = int(file_size)
            response = requests.get(url, headers=headers, stream=True)
            # 一块文件的大小
            # chunk_size = 1024
            chunk_size = int(file_size/100)+1
            bar = tqdm(total=file_size, desc=f'下载文件 {file_name}')
            with open(file_name, mode='wb') as f:
                # 写入分块文件
                for chunk in response.iter_content(chunk_size=chunk_size):
                    f.write(chunk)
                    bar.update(chunk_size)
            # 关闭进度条
            bar.close()
        else:
            # 未获取到文件大小，采用stream方法下载
            # download_from_url(url,file_name)
            download_by_stream(url,file_name)



import numpy as np
from netCDF4 import Dataset,date2num,num2date
import time
from datetime import datetime, timedelta

def creatspinc(value, data_vars, lats, lons, starttime, filename, resolution):

    gridspi = Dataset(filename, 'w', format='NETCDF4')

    # dimensions
    gridspi.createDimension('time', value[0].shape[0])
    gridspi.createDimension('lat', value[0].shape[2])   #len(lat)
    gridspi.createDimension('lon', value[0].shape[1])

    # Create coordinate variables for dimensions
    times = gridspi.createVariable('time', np.float64, ('time',))
    latitudes = gridspi.createVariable('lat', np.float32, ('lat',))
    longitudes = gridspi.createVariable('lon', np.float32, ('lon',))

    # Create the actual variable
    for var,attr in data_vars.items():
        gridspi.createVariable(var, np.float32, ('time', 'lon', 'lat',))

    # Global Attributes
    gridspi.description = 'var'
    gridspi.history = 'Created ' + time.ctime(time.time())
    gridspi.source = 'netCDF4 python module tutorial'

    # Variable Attributes
    latitudes.units = 'degree_north'
    longitudes.units = 'degree_east'
    times.units = 'days since 1970-01-01 00:00:00'
    times.calendar = 'gregorian'

    # data
    latitudes[:] = lats
    longitudes[:] = lons
    
    # Fill in times
    dates = []
    if resolution == 'daily':
        for n in range(value[0].shape[0]):
            dates.append(starttime + n)
        times[:] = dates[:]
    
    elif resolution == '6-hourly':
        # for n in range(value[0].shape[0]):
        #     dates.append(starttime + (n+1) * np.timedelta64(6, 'h'))
        
        for n in range(value[0].shape[0]):
            dates.append(starttime + (n+1) * timedelta(hours=6))

        times[:] = date2num(dates, units = times.units,calendar = times.calendar)
        # print 'time values (in units %s): ' % times.units +'\n', times[:]
        dates = num2date(times[:], units=times.units, calendar=times.calendar)
        
    # Fill in values  
    i = 0
    for var,attr in data_vars.items():
        gridspi.variables[var].long_name = attr['long_name']
        gridspi.variables[var].units = attr['units']
        gridspi.variables[var][:]= value[i][:]
        i = i + 1

    gridspi.close()
    

def regen_box(bbox, resolution, offset):
    
    lx = bbox[0]
    rx = bbox[2]
    LLON = round(int(lx) + resolution * int((lx - int(lx)) / resolution + 0.5) + offset * (int(lx * 10) / 10 + offset - lx) / abs(int(lx * 10) // 10 + offset - lx + 0.0000001), 3)
    RLON = round(int(rx) + resolution * int((rx - int(rx)) / resolution + 0.5) - offset * (int(rx * 10) / 10 + offset - rx) / abs(int(rx * 10) // 10 + offset - rx + 0.0000001), 3)
    
    by = bbox[1]
    ty = bbox[3]
    BLAT = round(int(by) + resolution * int((by - int(by)) / resolution + 0.5) + offset * (int(by * 10) / 10 + offset - by) / abs(int(by * 10) // 10 + offset - by + 0.0000001), 3)
    TLAT = round(int(ty) + resolution * int((ty - int(ty)) / resolution + 0.5) - offset * (int(ty * 10) / 10 + offset - ty) / abs(int(ty * 10) // 10 + offset - ty + 0.0000001), 3)
    
    # print(LLON,BLAT,RLON,TLAT)
    
    return [LLON,BLAT,RLON,TLAT]